using GraphicsManager.Enums;
using GraphicsManager.Interfaces;
using OpenTK.Mathematics;
using OpenTK.Windowing.Common;
using OpenTK.Windowing.Common.Input;
using OpenTK.Windowing.GraphicsLibraryFramework;

namespace GraphicsManager.Objects.Core;

public abstract class RenderObjectCore
{
    public event Func<RenderObjectCore, Task>? Clicked;
    public event Func<RenderObjectCore, string[], Task>? FilesDroped;
    public event Func<RenderObjectCore, Task>? MouseEnter;
    public event Func<RenderObjectCore, Task>? MouseLeave;
    public event Func<RenderObjectCore, Task>? SizeChanged;
    public event Func<RenderObjectCore, Task>? WindowLoaded;
    
    internal void SendFilesEvent(string[] Files)
    {
        if (FilesDroped is not null) _ = FilesDroped.Invoke(this, Files);
    }
    
    public abstract void DrawSelf();

    public ObjectAnchor Anchor { get; set; } = ObjectAnchor.Left | ObjectAnchor.Top;
    public bool Loaded { get; protected set; }
    public bool AllowHoverFromBehind
    {
        get => ahfb;
        set
        {
            ahfb = value;
            if (!value && MouseInside && Window is not null && Window.HoveringControl != this)
            {
                mi = false;
                MouseLeave?.Invoke(this);
            }
        }
    }
    public bool BlockDraw
    {
        get
        {
            if (Parent is null) return bd != 0;
            return Parent.BlockDraw || bd != 0;
        }
    }

    public BetterContextMenu? ContextMenu { get; set; } = null;
    public Vector2i Distance { get; internal set; }
    public bool IgnoreHover { get; set; }
    internal virtual void NotifiNotVisible()
    {
        IsVisible = false;
    }
    public bool IsVisible
    {
        get => iv && Visible;
        protected set => iv = value;
    }
    public virtual Vector4i Margins { get; set; } = new();
    public virtual MouseCursor HoverMouse { get; set; } = MouseCursor.Default;
    public bool MouseInside
    {
        get => mi;
        internal set
        {
            if (Window is null) return;
            if (AllowHoverFromBehind)
            {
                mi = value;
                if (value) MouseEnter?.Invoke(this);
                else MouseLeave?.Invoke(this);
            }
            else
            {
                if (value && Window.HoveringControl == this) MouseEnter?.Invoke(this);
                if (!value && Window.HoveringControl != this) MouseLeave?.Invoke(this);
            }
        }
    }
    public IParent? Parent { get; internal set; }
    public virtual Vector2i Location
    {
        get => loc_;
        set => SetLocation(value.X, value.Y);
    }
    public List<SubHitBox> SubHitBoxes { get; } = new();
    public object? Tag { get; set; }
    public virtual bool Visible
    {
        get => _Visible;
        set
        {
            _Visible = value;
            if (value) TryDraw();
            else Parent?.TryDraw();
        }
    }
    public virtual Vector2i Size
    {
        get => size_;
        set => SetSize(value.X, value.Y);
    }
    public IWindow? Window { get; internal set; }

    private bool ahfb = false;
    private bool iv;
    private bool mi = false;
    private bool _Visible = true;
    protected int bd = 0;
    private readonly object BD_LOCK = new();
    protected Vector2i loc_ = new(), size_ = new();

    public virtual void Focus() { }
    public virtual void UnFocus() { }

    public void ForceDistanceUpdate()
    {
        if (Parent is null) return;
        ForceDistanceUpdate(Parent);
    }
    public void ForceDistanceUpdate(IParent parent)
    {
        Distance = parent is IWindow w
            ? new Vector2i(w.CS.X - Size.X - Location.X, w.CS.Y - Size.Y - Location.Y)
            : new Vector2i(parent.Size.X - Size.X - Location.X, parent.Size.Y - Size.Y - Location.Y);
    }

    public abstract void TryDraw();
    public void AddBlockDraw()
    {
        lock (BD_LOCK) bd++;
    }
    public void RemoveBlockDraw()
    {
        lock (BD_LOCK) bd--;
    }

    public virtual void SetLocation(int xy) => SetLocation(xy, xy);
    public virtual void SetLocation(int x, int y)
    {
        loc_.X = x;
        loc_.Y = y;
        UpdateOpenGLCords(true);
    }

    public virtual void SetSize(int wh,
        bool UpdateOpenGLCords = true,
        bool TrigerSizeEvent = true,
        bool TrigerParentSizeEvent = true,
        bool TryDraw = true,
        bool UpdateDistance = false) => SetSize(wh, wh, UpdateOpenGLCords, TrigerSizeEvent, TrigerParentSizeEvent, TryDraw, UpdateDistance);
    public virtual void SetSize(int w, int h,
        bool UpdateOpenGLCords = true,
        bool TrigerSizeEvent = true,
        bool TrigerParentSizeEvent = true,
        bool TryDraw = true,
        bool UpdateDistance = false)
    {
        size_.X = w;
        size_.Y = h;
        if (UpdateOpenGLCords) this.UpdateOpenGLCords(TryDraw);
        else
        {
            if (TryDraw) this.TryDraw();
        }
        if (TrigerSizeEvent) SizeChanged?.Invoke(this);
        if (Parent is null) return;
        if (UpdateDistance) ForceDistanceUpdate(Parent);
        if (TrigerParentSizeEvent) Parent.ReportSizeUpdate(this);
        
    }

    public abstract void UpdateOpenGLCords(bool draw = false);
    internal abstract void LoadToParent(IParent Parent, IWindow Window);

    protected virtual void CallWindowLoaded()
    {
        if (WindowLoaded is not null) WindowLoaded.Invoke(this);
        Window!.MouseDown += Window_MouseDown;
    }
    
    public virtual void TrigerClick()
    {
        if (Clicked is not null) Clicked.Invoke(this);
    }
    
    protected void Window_MouseDown(OpenTK.Windowing.Common.MouseButtonEventArgs e)
    {
        if (MouseInside && e.Button == MouseButton.Button1 && Clicked is not null) _ = Clicked.Invoke(this);
        if (MouseInside && e.Button == MouseButton.Button2 && ContextMenu is not null && Window!.HoveringControl == this) ContextMenu.ShowContext(Window!);
        if (e.Button != MouseButton.Button2 && ContextMenu is not null && ContextMenu.Visible) ContextMenu.HideContext(Window!);
    }

    internal abstract void Clean();
    
    protected Vector2 laf = new(), saf = new();

    public Vector2 LocationAsFloat { get { return laf; } }
    public Vector2 SizeAsFloat { get { return saf; } }
    
    public virtual void SendClipEvent(string ClipString)
    {
		
    }
    
    public virtual void SendKeyEvent(KeyboardKeyEventArgs KeyArgs)
    {
		
    }
}
